package org.vraptor.component;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.vraptor.annotations.Component;
import org.vraptor.annotations.Destroy;
import org.vraptor.annotations.In;
import org.vraptor.annotations.Logic;
import org.vraptor.annotations.Out;
import org.vraptor.annotations.Read;
import org.vraptor.interceptor.InterceptorClass;
import org.vraptor.interceptor.InterceptorListFactory;
import org.vraptor.reflection.ReflectionUtil;
import org.vraptor.scope.ScopeType;

/**
 * A component class factory
 * 
 * @author Guilherme Silveira
 */
public class ComponentTypeFactory {

	private static final Logger LOG = Logger
			.getLogger(ComponentTypeFactory.class);

	public static final InterceptorListFactory INTERCEPTOR_LIST_FACTORY = new InterceptorListFactory();

	private final LogicMethodFactory factory = new LogicMethodFactory();

	/**
	 * Creates a component class dealer
	 * 
	 * @param type
	 *            the class to load
	 * @return
	 * @throws InvalidComponentException
	 */
	public ComponentType getComponentClass(Class<?> type)
			throws InvalidComponentException {

		LOG.debug("Reading component type " + type.getName());
		String componentName = getComponentName(type);

		if(!type.isAnnotationPresent(Component.class)) {
			LOG.warn("Deprecated: registering a component without @Component annotation.");
		}
		
		LOG.info("Component found: " + componentName + " from "
				+ type.getName());

		ScopeType scope = getScope(type);

		Constructor[] constructors = type.getConstructors();
		if (constructors.length != 1) {
			LOG.error("Component constructors found: " + constructors);
			throw new InvalidComponentException(
					type.getName()
							+ " component has more than one constructor. This is not desirable as it may reflect optional arguments and create complex attributes.");
		}
		ComponentConstructor constructor = new ComponentConstructor(
				constructors[0]);
		LOG.debug("Registered constructor: " + constructor);

		// read logics
		Map<String, DefaultLogicMethod> actions = new HashMap<String, DefaultLogicMethod>();
		for (Method m : type.getMethods()) {
			if (!Modifier.isPublic(m.getModifiers())
					|| m.getDeclaringClass().equals(Object.class)
					|| m.getName().startsWith("validate")) {
				continue;
			}
			Logic annotation = m.getAnnotation(Logic.class);
			generateLogics(actions, m, annotation, type);
		}

		List<InterceptorClass> interceptors = INTERCEPTOR_LIST_FACTORY
				.getInterceptors(type);

		// read fields
		List<FieldAnnotation<In>> inAnnotations = ReflectionUtil
				.readAnnotations(type, In.class);
		List<FieldAnnotation<Out>> outAnnotations = ReflectionUtil
				.readAnnotations(type, Out.class);
		List<FieldAnnotation<Read>> readAnnotations = ReflectionUtil
				.readAnnotations(type, Read.class);

		// destroy method
		String destroyLogicName = getDestroyLogicName(type);

		LOG.debug("Component clazz " + type.getName() + " read");

		return new DefaultComponentType(type, componentName, scope,
				constructor, actions, inAnnotations, outAnnotations,
				readAnnotations, interceptors, destroyLogicName);

	}

	String getDestroyLogicName(Class<?> type) {
		String destroyLogicName = "destroy";
		for (Method m : type.getMethods()) {
			if (m.isAnnotationPresent(Destroy.class)) {
				destroyLogicName = m.getName();
			}
		}
		return destroyLogicName;
	}

	ScopeType getScope(Class<?> type) {
		return type.isAnnotationPresent(Component.class) ? type.getAnnotation(
				Component.class).scope() : ScopeType.LOGIC;
	}

	String getComponentName(Class<?> type) {
		String componentName = type.getSimpleName();

		if (type.isAnnotationPresent(Component.class)) {
			Component ann = ((Component) type.getAnnotation(Component.class));
			if (!ann.value().equals(Component.EMPTY_NAME)) {
				componentName = ann.value();
			}
		} 
		return componentName;
	}

	/**
	 * Generate logics for an specific method and registers them
	 * 
	 * @param actions
	 *            actions
	 * @param method
	 *            method
	 * @param annotation
	 *            logic annotation
	 * @param componentClass
	 * @throws InvalidComponentException
	 */
	private void generateLogics(Map<String, DefaultLogicMethod> actions,
			Method method, Logic annotation, Class<?> componentClass)
			throws InvalidComponentException {
		if (annotation == null) {
			register(actions, factory.create(method.getName(), componentClass,
					method));
			return;
		}
		for (String name : annotation.value()) {
			register(actions, factory.create(name, componentClass, method));
		}
	}

	/**
	 * @param actions
	 * @param action
	 */
	private static void register(Map<String, DefaultLogicMethod> actions,
			DefaultLogicMethod action) {
		LOG.info("Registering logic " + action.getName());
		actions.put(action.getName(), action);
	}

}
