package org.vraptor.component;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.vraptor.LogicRequest;
import org.vraptor.annotations.In;
import org.vraptor.annotations.Out;
import org.vraptor.annotations.Read;
import org.vraptor.interceptor.InterceptorClass;
import org.vraptor.introspector.FieldReadParameter;
import org.vraptor.scope.ScopeType;

/**
 * Metadata information about a component class. This class is used to dealing
 * with component classes.
 * 
 * @author Guilherme Silveira
 */
public class DefaultComponentType implements ComponentType {

	private static final Logger logger = Logger
			.getLogger(DefaultComponentType.class);

	private final Map<String, DefaultLogicMethod> actions;

	private final List<FieldAnnotation<In>> inAnnotations;

	private final List<FieldAnnotation<Out>> outAnnotations;

	private final List<FieldReadParameter> readParameters = new ArrayList<FieldReadParameter>();

	private final List<InterceptorClass> interceptors;

	private final Class<?> clazz;

	private final String name;

	private final ComponentConstructor constructor;

	private final ScopeType scope;

	private final String destroyLogicName;

	<T> DefaultComponentType(Class<T> clazz, String name, ScopeType scope,
			ComponentConstructor constructor,
			Map<String, DefaultLogicMethod> actions,
			List<FieldAnnotation<In>> inAnnotations,
			List<FieldAnnotation<Out>> outAnnotations,
			List<FieldAnnotation<Read>> readAnnotations,
			List<InterceptorClass> interceptors, String destroyLogicName) {
		this.name = name;
		this.actions = actions;
		this.scope = scope;
		this.inAnnotations = inAnnotations;
		this.outAnnotations = outAnnotations;
		this.interceptors = interceptors;
		this.clazz = clazz;
		this.constructor = constructor;
		for (FieldAnnotation<Read> info : readAnnotations) {
			this.readParameters.add(new FieldReadParameter(info));
		}
		this.destroyLogicName = destroyLogicName;
	}

	public Object newInstance(LogicRequest request)
			throws ComponentInstantiationException {
		logger.debug("Instantiating class " + this.clazz.getName());
		return this.constructor.newInstance(request);
	}

	public String getName() {
		return this.name;
	}

	public LogicMethod getLogic(String key) throws LogicNotFoundException {
		if (!this.actions.containsKey(key)) {
			throw new LogicNotFoundException("Unable to find action " + key);
		}
		return this.actions.get(key);
	}

	public List<InterceptorClass> getInterceptors() {
		return this.interceptors;
	}

	public List<FieldAnnotation<In>> getInAnnotations() {
		return this.inAnnotations;
	}

	public List<FieldAnnotation<Out>> getOutAnnotations() {
		return this.outAnnotations;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.vraptor.component.ComponentType#getReadParameters()
	 */
	public List<FieldReadParameter> getReadParameters() {
		return this.readParameters;
	}

	public Collection<LogicMethod> getLogics() {
		return new ArrayList<LogicMethod>(this.actions.values());
	}

	public Class getComponentClass() {
		return this.clazz;
	}

	public ScopeType getScope() {
		return this.scope;
	}

	public String getKey() {
		return "component" + Character.toUpperCase(getComponentClass().getSimpleName().charAt(0))
				+ getComponentClass().getSimpleName().substring(1);
	}

	public String getDestroyLogicName() {
		return this.destroyLogicName;
	}

}
