package org.vraptor.webapp;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Set;

import javax.servlet.ServletContext;

import org.apache.log4j.Logger;
import org.vraptor.VRaptorServlet;
import org.vraptor.annotations.Component;
import org.vraptor.component.ComponentManager;
import org.vraptor.component.InvalidComponentException;
import org.vraptor.config.ConfigException;
import org.vraptor.config.XMLConfigurationLoader;
import org.vraptor.converter.ConverterManager;
import org.vraptor.converter.SimpleConverterManager;
import org.vraptor.factory.FactoryManager;
import org.vraptor.factory.SimpleFactoryManager;
import org.vraptor.plugin.DefaultPluginManager;
import org.vraptor.scope.ApplicationContext;
import org.vraptor.scope.SimpleApplicationContext;
import org.vraptor.url.DefaultURLManager;
import org.vraptor.url.URLManager;
import org.vraptor.view.RegexViewManager;
import org.vraptor.view.ViewManager;

/**
 * A simple web application configuration. It uses the default url manager.
 * 
 * @author Guilherme Silveira
 * @author Paulo Silveira
 */
public class DefaultWebApplication implements WebApplication {

	private static final String VRAPTOR_XML_LOCATION = "/vraptor.xml";

	private static final String JSP_REGEX = "/$1/$2.$3.jsp";

	private static final Logger LOG = Logger.getLogger(DefaultWebApplication.class);

	private static final String WEBINF_CLASSES_DIRECTORY = "/WEB-INF/classes/";

	private static final String DOT_CLASS = ".class";

	private ConverterManager converterManager;

	private URLManager urlManager;

	private ComponentManager componentManager;

	private FactoryManager factoryManager;

	private ViewManager viewManager;

	private PluginManager pluginManager;

	private ApplicationContext applicationContext;

	public DefaultWebApplication(ServletContext context) {
		applicationContext = new SimpleApplicationContext(context);
		urlManager = new DefaultURLManager();
		componentManager = new DefaultComponentManager();
		factoryManager = new SimpleFactoryManager();
		viewManager = new RegexViewManager(JSP_REGEX);
		converterManager = new SimpleConverterManager();
		pluginManager = new DefaultPluginManager();
	}

	public ApplicationContext getApplicationContext() {
		return applicationContext;
	}

	public URLManager getURLManager() {
		return urlManager;
	}

	/**
	 * Reads the vraptor.xml file. If not found, search for component classes
	 * class loading all of them.
	 * 
	 * Inits this configuration: overrides the viewmanager with the
	 * views.properties file
	 * 
	 * @throws ConfigException
	 * 
	 * @see org.vraptor.webapp.WebApplication#init()
	 */
	public void init() throws ConfigException {
		// reads the xml file
		InputStream file = getXMLFile();
		if (file == null) {
			LOG.info("No vraptor.xml found, loading *all* WEB-INF classes searching for components");
			try {
				readComponentsLoadingClasses(WEBINF_CLASSES_DIRECTORY);
			} catch (InvalidComponentException e) {
				throw new ConfigException("cannot load classes", e);
			}
		} else {
			try {
				new XMLConfigurationLoader().readManagers(file, this);
				file.close();
				file = getXMLFile();
				new XMLConfigurationLoader().readComponents(file, this);
				file.close();
			} catch (IOException e) {
				throw new ConfigException(e.getMessage(), e);
			}
		}

		// reads overriden views.properties
		InputStream viewProperties = DefaultWebApplication.class.getResourceAsStream("/views.properties");
		if (viewProperties != null) {
			try {
				this.viewManager = new ViewsPropertiesReader().overrideViews(viewManager, viewProperties);
			} catch (IOException e) {
				LOG.warn("Error reading views.properties", e);
			}
		}
	}

	/**
	 * Reads all .class files from <code>directory</code> looking for
	 * annotated vraptor classes (@Component).
	 * 
	 * @param directory
	 * @throws InvalidComponentException
	 */
	@SuppressWarnings("unchecked")
	private void readComponentsLoadingClasses(String directory) throws InvalidComponentException {
		// TODO: better code? shale does something similar
		// TODO: check jetty compatibility
		Set<String> items = applicationContext.getResourcePaths(directory);
		for (String item : items) {
			File file = new File(applicationContext.getRealPath(item));
			if (file.isDirectory()) {
				readComponentsLoadingClasses(item);
			} else if (item.endsWith(DOT_CLASS)) {
				String className = parseClassName(WEBINF_CLASSES_DIRECTORY, item);
				if (classIsComponent(className)) {
					LOG.debug(String.format("class found in dir %s: %s", directory, className));
					this.componentManager.register(className);
				}
			}
		}
	}

	/**
	 * Checks if className is a valid Component (annotated)
	 * 
	 * @param className
	 * @return
	 * @throws IllegalArgumentException
	 *             if className cannot be loaded
	 */
	private static boolean classIsComponent(String className) {
		try {
			return Class.forName(className).isAnnotationPresent(Component.class);
		} catch (ClassNotFoundException e) {
			throw new IllegalArgumentException(String.format("cant check if %s is a class, unable to load it",
					className), e);
		}
	}

	/**
	 * Parses a String like
	 * <code>/WEB-INF/classes/br/com/caelum/Test.class to</code>
	 * <code>br.com.caelum.Test.class</code>
	 * 
	 * @param directory
	 * @param name
	 * @return
	 */
	public static String parseClassName(String directory, String name) {
		if (!name.startsWith(directory)) {
			throw new IllegalArgumentException();
		}
		name = name.substring(directory.length(), name.length() - DOT_CLASS.length());
		name = name.replaceAll("/|\\\\", ".");
		return name;
	}

	private InputStream getXMLFile() {
		InputStream file = VRaptorServlet.class.getResourceAsStream(VRAPTOR_XML_LOCATION);
		return file;
	}

	public FactoryManager getFactoryManager() {
		return this.factoryManager;
	}

	public ComponentManager getComponentManager() {
		return this.componentManager;
	}

	public ViewManager getViewManager() {
		return this.viewManager;
	}

	public ConverterManager getConverterManager() {
		return this.converterManager;
	}

	public void setViewManager(ViewManager viewManager) {
		this.viewManager = viewManager;
	}

	public PluginManager getPluginManager() {
		return this.pluginManager;
	}

}
